/***
 * @Copyright 2022-2023 SPACEMIT. All rights reserved.
 * @Use of this source code is governed by a BSD-style license
 * @that can be found in the LICENSE file.
 * @
 * @Author: David(qiang.fu@spacemit.com)
 * @Date: 2023-08-30 15:31:42
 * @LastEditTime: 2023-09-20 14:03:32
 * @Description:
 */

#ifndef LIBV4L_STMPP_H
#define LIBV4L_STMPP_H

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <pthread.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/queue.h>
#include <sys/syscall.h>
#include <unistd.h>
#include <vdec.h>
#include <venc.h>

#include "config.h"
#include "linux/videodev2.h"

#define LIBV4L_STMPP_VERSION "0.0.15"

extern int stmpp_log_level;

#define gettid() syscall(SYS_gettid)

#define LOG(fmt, ...)                                                       \
  do {                                                                      \
    struct timeval tv;                                                      \
    gettimeofday(&tv, NULL);                                                \
    printf("[%03ld.%03ld] [STMPP] [%ld] %s(%d): " fmt, tv.tv_sec % 1000,    \
           tv.tv_usec / 1000, gettid(), __func__, __LINE__, ##__VA_ARGS__); \
    fflush(stdout);                                                         \
  } while (0)

#define LOGV(level, fmt, ...)                              \
  do {                                                     \
    if (stmpp_log_level >= level) LOG(fmt, ##__VA_ARGS__); \
  } while (0)

#define LOGE(fmt, ...) LOG("ERR: " fmt, ##__VA_ARGS__)

#define RETURN_ERR(err, ret)       \
  {                                \
    errno = err;                   \
    LOGV(2, "errno: %d\n", errno); \
    return ret;                    \
  }

#define ENTER() LOGV(5, "ctx(%p): ENTER\n", (void *)ctx)
#define LEAVE() LOGV(5, "ctx(%p): LEAVE\n", (void *)ctx)

#define min(x, y) ((x) < (y) ? (x) : (y))
#define max(x, y) ((x) > (y) ? (x) : (y))

/* From kernel's linux/kernel.h */
#define ARRAY_SIZE(arr) (sizeof(arr) / sizeof((arr)[0]))

#define clamp(val, lo, hi) min((typeof(val))max(val, lo), hi)

#define __round_mask(x, y) ((__typeof__(x))((y)-1))
#define round_up(x, y) ((((x)-1) | __round_mask(x, y)) + 1)

/* From kernel's linux/stddef.h */
#ifndef offsetof
#define offsetof(TYPE, MEMBER) ((size_t) & ((TYPE *)0)->MEMBER)
#endif

/* From kernel's v4l2-core/v4l2-ioctl.c */
#define CLEAR_AFTER_FIELD(p, field)                                           \
  memset((uint8_t *)(p) + offsetof(typeof(*(p)), field) + sizeof((p)->field), \
         0, sizeof(*(p)) - offsetof(typeof(*(p)), field) - sizeof((p)->field))

#define MPP_VIDEO_CodingNone 0  // MPP_VIDEO_CodingUnused

#define STMPP_MB_DIM 16
#define STMPP_SB_DIM 64

#define STMPP_MAX_PLANE 3

#define STMPP_MEM_OFFSET(type, index) ((int64_t)((type) << 16 | (index)))
#define STMPP_MEM_OFFSET_TYPE(offset) (int)((offset) >> 16)
#define STMPP_MEM_OFFSET_INDEX(offset) (int)((offset) & ((1 << 16) - 1))

#define STMPP_HAS_FORMAT(ctx, format)                           \
  (!((format)->type != MPP_VIDEO_CodingNone && (ctx)->codecs && \
     !strstr((ctx)->codecs, (format)->name)))

/**
 * struct stmpp_fmt - Information about mpp supported format
 * @name:		Format's name.
 * @fourcc:		Format's forcc.
 * @num_planes:		Number of planes.
 * @type:		Format's mpp coding type.
 * @format:		Format's mpp frame format.
 * @depth:		Format's pixel depth.
 * @frmsize:		V4L2 frmsize_stepwise.
 */
struct stmpp_fmt {
  char *name;
  uint32_t fourcc;
  int num_planes;
  MppCodingType type;
  MppPixelFormat format;
  uint8_t depth[VIDEO_MAX_PLANES];
  struct v4l2_frmsize_stepwise frmsize;
};

/**
 * enum stmpp_buffer_flag - Flags of stmpp buffer
 * @ERROR:		Something wrong in the buffer.
 * @LOCKED:		Buffer been locked from mpp buffer group.
 * @EXPORTED:		Buffer been exported to userspace.
 * @QUEUED:		Buffer been queued.
 * @PENDING:		Buffer is in pending queue.
 * @AVAILABLE:		Buffer is in available queue.
 */
enum stmpp_buffer_flag {
  STMPP_BUFFER_ERROR = 1 << 0,
  STMPP_BUFFER_LOCKED = 1 << 1,
  STMPP_BUFFER_EXPORTED = 1 << 2,
  STMPP_BUFFER_QUEUED = 1 << 3,
  STMPP_BUFFER_PENDING = 1 << 4,
  STMPP_BUFFER_AVAILABLE = 1 << 5,
  STMPP_BUFFER_KEYFRAME = 1 << 6,
};

/**
 * struct stmpp_buffer - Information about mpp buffer
 * @entry:		Queue entry.
 * @stmpp_buf:		Handle of mpp buffer.
 * @index:		Buffer's index.
 * @fd:			Buffer's dma fd.
 * @timestamp:		Buffer's timestamp.
 * @bytesused:		Number of bytes occupied by data in the buffer.
 * @length:		Buffer's length(planes).
 * @size:		Buffer's size.
 * @flags:		Buffer's flags.
 * @planes:		Buffer's planes info.
 */
struct stmpp_buffer {
  TAILQ_ENTRY(stmpp_buffer) entry;
  int is_frame;
  MppFrame *stmpp_frame;
  MppPacket *stmpp_packet;
  int is_free;

  int index;

  int fd;
  uint64_t timestamp;
  uint32_t bytesused;
  uint32_t length;
  uint32_t flags;
  uint32_t size;
  uint32_t type;

  struct {
    unsigned long userptr;
    int fd;
    uint32_t data_offset;
    uint32_t bytesused;
    uint32_t plane_size; /* bytesused - data_offset */
    uint32_t length;
  } planes[STMPP_MAX_PLANE];
};

TAILQ_HEAD(stmpp_buf_head, stmpp_buffer);

/**
 * struct stmpp_buf_head - Information about mpp buffer queue
 * @memory:		V4L2 memory type.
 * @streaming:		The queue is streaming.
 * @internal_group:	Handle of mpp internal buffer group.
 * @external_group:	Handle of mpp external buffer group.
 * @buffers:		List of buffers.
 * @num_buffers:	Number of buffers.
 * @avail_buffers:	Buffers ready to be dequeued.
 * @pending_buffers:	Pending buffers for mpp.
 * @queue_mutex:	Mutex for buffer lists.
 * @stmpp_format:	Mpp format.
 * @format:		V4L2 multi-plane format.
 */
struct stmpp_buf_queue {
  enum v4l2_memory memory;

  bool streaming;

  // MppBufferGroup internal_group;
  // MppBufferGroup external_group;
  struct stmpp_buffer *buffers;
  uint32_t num_buffers;

  struct stmpp_buf_head avail_buffers;
  struct stmpp_buf_head pending_buffers;
  pthread_mutex_t queue_mutex;

  const struct stmpp_fmt *stmpp_format;
  struct v4l2_pix_format_mplane format;
};

/**
 * struct stmpp_context - Context data
 * @formats:		Supported formats.
 * @num_formats:	Number of formats.
 * @is_decoder:		Is decoder mode.
 * @nonblock:		Nonblock mode.
 * @eventfd:		File descriptor of eventfd.
 * @avail_buffers:	Buffers ready to be dequeued.
 * @pending_buffers:	Pending buffers for mpp.
 * @mpp:		Handler of mpp context.
 * @mpi:		Handler of mpp api.
 * @output:		Output queue.
 * @capture:		Capture queue.
 * @ioctl_mutex:	Mutex.
 * @frames:		Number of frames reported.
 * @last_fps_time:	The last time to count fps.
 * @data:		Private data.
 */
struct stmpp_context {
  struct stmpp_fmt *formats;
  uint32_t num_formats;

  bool is_decoder;
  bool nonblock;
  int eventfd;

  // MppCtx mpp;
  // MppApi *mpi;

  struct stmpp_buf_queue output;
  struct stmpp_buf_queue capture;

  pthread_mutex_t ioctl_mutex;

  uint64_t frames;
  uint64_t last_fps_time;

  unsigned int max_width;
  unsigned int max_height;
  char *codecs;

  void *data;
  MppFrame *frame[64];
};

static inline struct stmpp_buf_queue *stmpp_get_queue(struct stmpp_context *ctx,
                                                      enum v4l2_buf_type type) {
  LOGV(4, "type = %d\n", type);

  switch (type) {
    case V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE:
      return &ctx->capture;
    case V4L2_BUF_TYPE_VIDEO_OUTPUT_MPLANE:
      return &ctx->output;
    default:
      LOGE("invalid buf type\n");
      RETURN_ERR(EINVAL, NULL);
  }
}

static inline const char *stmpp_cmd2str(unsigned long cmd) {
#define STMPP_CMD2STR(cmd) \
  case cmd:                \
    return #cmd

  switch (cmd) {
    STMPP_CMD2STR(VIDIOC_QUERYCAP);
    STMPP_CMD2STR(VIDIOC_ENUM_FMT);
    STMPP_CMD2STR(VIDIOC_G_FMT);
    STMPP_CMD2STR(VIDIOC_S_FMT);
    STMPP_CMD2STR(VIDIOC_REQBUFS);
    STMPP_CMD2STR(VIDIOC_QUERYBUF);
    STMPP_CMD2STR(VIDIOC_G_FBUF);
    STMPP_CMD2STR(VIDIOC_S_FBUF);
    STMPP_CMD2STR(VIDIOC_OVERLAY);
    STMPP_CMD2STR(VIDIOC_QBUF);
    STMPP_CMD2STR(VIDIOC_EXPBUF);
    STMPP_CMD2STR(VIDIOC_DQBUF);
    STMPP_CMD2STR(VIDIOC_STREAMON);
    STMPP_CMD2STR(VIDIOC_STREAMOFF);
    STMPP_CMD2STR(VIDIOC_G_PARM);
    STMPP_CMD2STR(VIDIOC_S_PARM);
    STMPP_CMD2STR(VIDIOC_G_STD);
    STMPP_CMD2STR(VIDIOC_S_STD);
    STMPP_CMD2STR(VIDIOC_ENUMSTD);
    STMPP_CMD2STR(VIDIOC_ENUMINPUT);
    STMPP_CMD2STR(VIDIOC_G_CTRL);
    STMPP_CMD2STR(VIDIOC_S_CTRL);
    STMPP_CMD2STR(VIDIOC_G_TUNER);
    STMPP_CMD2STR(VIDIOC_S_TUNER);
    STMPP_CMD2STR(VIDIOC_G_AUDIO);
    STMPP_CMD2STR(VIDIOC_S_AUDIO);
    STMPP_CMD2STR(VIDIOC_QUERYCTRL);
    STMPP_CMD2STR(VIDIOC_QUERYMENU);
    STMPP_CMD2STR(VIDIOC_G_INPUT);
    STMPP_CMD2STR(VIDIOC_S_INPUT);
    STMPP_CMD2STR(VIDIOC_G_EDID);
    STMPP_CMD2STR(VIDIOC_S_EDID);
    STMPP_CMD2STR(VIDIOC_G_OUTPUT);
    STMPP_CMD2STR(VIDIOC_S_OUTPUT);
    STMPP_CMD2STR(VIDIOC_ENUMOUTPUT);
    STMPP_CMD2STR(VIDIOC_G_AUDOUT);
    STMPP_CMD2STR(VIDIOC_S_AUDOUT);
    STMPP_CMD2STR(VIDIOC_G_MODULATOR);
    STMPP_CMD2STR(VIDIOC_S_MODULATOR);
    STMPP_CMD2STR(VIDIOC_G_FREQUENCY);
    STMPP_CMD2STR(VIDIOC_S_FREQUENCY);
    STMPP_CMD2STR(VIDIOC_CROPCAP);
    STMPP_CMD2STR(VIDIOC_G_CROP);
    STMPP_CMD2STR(VIDIOC_S_CROP);
    STMPP_CMD2STR(VIDIOC_G_JPEGCOMP);
    STMPP_CMD2STR(VIDIOC_S_JPEGCOMP);
    STMPP_CMD2STR(VIDIOC_QUERYSTD);
    STMPP_CMD2STR(VIDIOC_TRY_FMT);
    STMPP_CMD2STR(VIDIOC_ENUMAUDIO);
    STMPP_CMD2STR(VIDIOC_ENUMAUDOUT);
    STMPP_CMD2STR(VIDIOC_G_PRIORITY);
    STMPP_CMD2STR(VIDIOC_S_PRIORITY);
    STMPP_CMD2STR(VIDIOC_G_SLICED_VBI_CAP);
    STMPP_CMD2STR(VIDIOC_LOG_STATUS);
    STMPP_CMD2STR(VIDIOC_G_EXT_CTRLS);
    STMPP_CMD2STR(VIDIOC_S_EXT_CTRLS);
    STMPP_CMD2STR(VIDIOC_TRY_EXT_CTRLS);
    STMPP_CMD2STR(VIDIOC_ENUM_FRAMESIZES);
    STMPP_CMD2STR(VIDIOC_ENUM_FRAMEINTERVALS);
    STMPP_CMD2STR(VIDIOC_G_ENC_INDEX);
    STMPP_CMD2STR(VIDIOC_ENCODER_CMD);
    STMPP_CMD2STR(VIDIOC_TRY_ENCODER_CMD);
    STMPP_CMD2STR(VIDIOC_DBG_S_REGISTER);
    STMPP_CMD2STR(VIDIOC_DBG_G_REGISTER);
    STMPP_CMD2STR(VIDIOC_S_HW_FREQ_SEEK);
    STMPP_CMD2STR(VIDIOC_S_DV_TIMINGS);
    STMPP_CMD2STR(VIDIOC_G_DV_TIMINGS);
    STMPP_CMD2STR(VIDIOC_DQEVENT);
    STMPP_CMD2STR(VIDIOC_SUBSCRIBE_EVENT);
    STMPP_CMD2STR(VIDIOC_UNSUBSCRIBE_EVENT);
    STMPP_CMD2STR(VIDIOC_CREATE_BUFS);
    STMPP_CMD2STR(VIDIOC_PREPARE_BUF);
    STMPP_CMD2STR(VIDIOC_G_SELECTION);
    STMPP_CMD2STR(VIDIOC_S_SELECTION);
    STMPP_CMD2STR(VIDIOC_DECODER_CMD);
    STMPP_CMD2STR(VIDIOC_TRY_DECODER_CMD);
    STMPP_CMD2STR(VIDIOC_ENUM_DV_TIMINGS);
    STMPP_CMD2STR(VIDIOC_QUERY_DV_TIMINGS);
    STMPP_CMD2STR(VIDIOC_DV_TIMINGS_CAP);
    STMPP_CMD2STR(VIDIOC_ENUM_FREQ_BANDS);
    STMPP_CMD2STR(VIDIOC_DBG_G_CHIP_INFO);
    STMPP_CMD2STR(VIDIOC_QUERY_EXT_CTRL);
    default:
      return "UNKNOWN";
  }
}

#define STMPP_BUFFER_FLAG_HELPER_GET(flag, name)                        \
  static inline bool stmpp_buffer_##name(struct stmpp_buffer *buffer) { \
    return !!(buffer->flags & flag);                                    \
  }

#define STMPP_BUFFER_FLAG_HELPER_SET(flag, name)                            \
  static inline void stmpp_buffer_set_##name(struct stmpp_buffer *buffer) { \
    LOGV(4, "buffer: %d type: %d\n", buffer->index, buffer->type);          \
    if (stmpp_buffer_##name(buffer))                                        \
      LOGE("buffer: %d type: %d is already " #name "\n", buffer->index,     \
           buffer->type);                                                   \
    buffer->flags |= flag;                                                  \
  }

#define STMPP_BUFFER_FLAG_HELPER_CLR(flag, name)                            \
  static inline void stmpp_buffer_clr_##name(struct stmpp_buffer *buffer) { \
    LOGV(4, "buffer: %d type: %d\n", buffer->index, buffer->type);          \
    if (!stmpp_buffer_##name(buffer))                                       \
      LOGE("buffer: %d type: %d is not " #name "\n", buffer->index,         \
           buffer->type);                                                   \
    buffer->flags &= ~flag;                                                 \
  }

#define STMPP_BUFFER_FLAG_HELPERS(flag, name) \
  STMPP_BUFFER_FLAG_HELPER_GET(flag, name)    \
  STMPP_BUFFER_FLAG_HELPER_SET(flag, name)    \
  STMPP_BUFFER_FLAG_HELPER_CLR(flag, name)

STMPP_BUFFER_FLAG_HELPERS(STMPP_BUFFER_ERROR, error)
STMPP_BUFFER_FLAG_HELPERS(STMPP_BUFFER_LOCKED, locked)
STMPP_BUFFER_FLAG_HELPERS(STMPP_BUFFER_EXPORTED, exported)
STMPP_BUFFER_FLAG_HELPERS(STMPP_BUFFER_QUEUED, queued)
STMPP_BUFFER_FLAG_HELPERS(STMPP_BUFFER_PENDING, pending)
STMPP_BUFFER_FLAG_HELPERS(STMPP_BUFFER_AVAILABLE, available)
STMPP_BUFFER_FLAG_HELPERS(STMPP_BUFFER_KEYFRAME, keyframe)

void stmpp_new_frame(struct stmpp_context *ctx);
int stmpp_update_poll_event(struct stmpp_context *ctx);
int stmpp_querycap(struct stmpp_context *ctx, struct v4l2_capability *cap);
int stmpp_enum_fmt(struct stmpp_context *ctx, struct v4l2_fmtdesc *f);
int stmpp_enum_framesizes(struct stmpp_context *ctx,
                          struct v4l2_frmsizeenum *fsize);
int stmpp_try_fmt(struct stmpp_context *ctx, struct v4l2_format *f);
int stmpp_s_fmt(struct stmpp_context *ctx, struct v4l2_format *f);
int stmpp_g_fmt(struct stmpp_context *ctx, struct v4l2_format *f);
int stmpp_reqbufs(struct stmpp_context *ctx,
                  struct v4l2_requestbuffers *reqbufs);
int stmpp_querybuf(struct stmpp_context *ctx, struct v4l2_buffer *buffer);
int stmpp_expbuf(struct stmpp_context *ctx, struct v4l2_exportbuffer *expbuf);
int stmpp_qbuf(struct stmpp_context *ctx, struct v4l2_buffer *buffer);
int stmpp_dqbuf(struct stmpp_context *ctx, struct v4l2_buffer *buffer);

/* Utils */

int stmpp_to_v4l2_buffer(struct stmpp_context *ctx,
                         struct stmpp_buffer *stmpp_buffer,
                         struct v4l2_buffer *buffer);
int stmpp_from_v4l2_buffer(struct stmpp_context *ctx,
                           struct v4l2_buffer *buffer,
                           struct stmpp_buffer *stmpp_buffer);

#endif  // LIBV4L_STMPP_H
