/*
 * Copyright 2022-2023 SPACEMIT. All rights reserved.
 * Use of this source code is governed by a BSD-style license
 * that can be found in the LICENSE file.
 *
 * @Author: David(qiang.fu@spacemit.com)
 * @Date: 2023-09-15 14:43:12
 * @LastEditTime: 2023-11-07 16:05:36
 * @Description:
 */

#include "libv4l-stmpp-dec.h"

#include <signal.h>

static struct stmpp_fmt stmpp_dec_fmts[] = {
    {
        .name = "4:2:0 1 plane Y/CbCr",
        .fourcc = V4L2_PIX_FMT_NV12,
        .num_planes = 1,
        .type = MPP_VIDEO_CodingNone,
        .format = PIXEL_FORMAT_NV12,
        .depth = {12},
    },
    {
        .name = "H.265",
        .fourcc = V4L2_PIX_FMT_HEVC,
        .num_planes = 1,
        .type = CODING_H265,
        //.format = MPP_FMT_BUTT,
        .frmsize =
            {
                .min_width = 48,
                .max_width = 3840,
                .step_width = STMPP_MB_DIM,
                .min_height = 48,
                .max_height = 2160,
                .step_height = STMPP_MB_DIM,
            },
    },
    {
        .name = "H.264",
        .fourcc = V4L2_PIX_FMT_H264,
        .num_planes = 1,
        .type = CODING_H264,
        //.format = MPP_FMT_BUTT,
        .frmsize =
            {
                .min_width = 48,
                .max_width = 3840,
                .step_width = STMPP_MB_DIM,
                .min_height = 48,
                .max_height = 2160,
                .step_height = STMPP_MB_DIM,
            },
    },
    {
        .name = "VP8",
        .fourcc = V4L2_PIX_FMT_VP8,
        .num_planes = 1,
        .type = CODING_VP8,
        //.format = MPP_FMT_BUTT,
        .frmsize =
            {
                .min_width = 48,
                .max_width = 3840,
                .step_width = STMPP_MB_DIM,
                .min_height = 48,
                .max_height = 2160,
                .step_height = STMPP_MB_DIM,
            },
    },
    {
        .name = "VP9",
        .fourcc = V4L2_PIX_FMT_VP9,
        .num_planes = 1,
        .type = CODING_VP9,
        //.format = MPP_FMT_BUTT,
        .frmsize =
            {
                .min_width = 48,
                .max_width = 3840,
                .step_width = STMPP_SB_DIM,
                .min_height = 48,
                .max_height = 2160,
                .step_height = STMPP_SB_DIM,
            },
    },
};

// Feed all available packets to mpp
static void stmpp_put_packets(struct stmpp_dec_context *dec) {
  struct stmpp_context *ctx = dec->ctx;
  struct stmpp_buffer *stmpp_buffer;

  int ret;
  bool is_eos;

  ENTER();

  pthread_mutex_lock(&ctx->output.queue_mutex);

  while (!TAILQ_EMPTY(&ctx->output.pending_buffers)) {
    stmpp_buffer = TAILQ_FIRST(&ctx->output.pending_buffers);

    PACKET_SetLength(stmpp_buffer->stmpp_packet, stmpp_buffer->bytesused);
    PACKET_SetPts(stmpp_buffer->stmpp_packet, stmpp_buffer->timestamp);
    // TODO: Support start/stop decode cmd
    // The chromium uses -2 as special flush timestamp.
    is_eos = stmpp_buffer->timestamp == (uint64_t)-2000000;
    if (is_eos) {
      PACKET_SetEos(stmpp_buffer->stmpp_packet, 1);
    } else {
      PACKET_SetEos(stmpp_buffer->stmpp_packet, 0);
    }

    // LOGE("stmpp_put_packets bytesused = %d\n", stmpp_buffer->bytesused);
    ret =
        VDEC_Decode(dec->pVdecCtx,
                    PACKET_GetBaseData(stmpp_buffer->stmpp_packet /*packet*/));

    if (ret != MPP_OK) {
      if (ret == MPP_DATAQUEUE_FULL || ret == MPP_POLL_FAILED) {
        // input unblock mode will go here, but now we use input block mode.
        LOGV(1, "Buffer full writing %d bytes to decoder\n",
             stmpp_buffer->bytesused);
      } else {
        LOGE("other error %d\n", ret);
      }

      usleep(1000);
      break;
    } else {
      LOGV(1, "Write %d bytes to decoder\n", stmpp_buffer->bytesused);
    }

    TAILQ_REMOVE(&ctx->output.pending_buffers, stmpp_buffer, entry);
    stmpp_buffer_clr_pending(stmpp_buffer);

    // Hold eos packet until eos frame received(flushed)
    if (is_eos) {
      LOGV(1, "hold eos packet: %d\n", stmpp_buffer->index);
      dec->eos_packet = stmpp_buffer;
      break;
    }

    LOGV(3, "put packet: %d(%" PRIu64 ") len=%d\n", stmpp_buffer->index,
         stmpp_buffer->timestamp, stmpp_buffer->bytesused);

    stmpp_buffer->bytesused = 0;

    LOGV(3, "return packet: %d\n", stmpp_buffer->index);

    TAILQ_INSERT_TAIL(&ctx->output.avail_buffers, stmpp_buffer, entry);
    stmpp_buffer_set_available(stmpp_buffer);
  }
  pthread_mutex_unlock(&ctx->output.queue_mutex);

  LEAVE();
}

// Feed all available frames to mpp
static void stmpp_put_frames(struct stmpp_dec_context *dec) {
  struct stmpp_context *ctx = dec->ctx;
  struct stmpp_buffer *stmpp_buffer;

  ENTER();

  pthread_mutex_lock(&ctx->capture.queue_mutex);
  while (!TAILQ_EMPTY(&ctx->capture.pending_buffers)) {
    stmpp_buffer = TAILQ_FIRST(&ctx->capture.pending_buffers);
    TAILQ_REMOVE(&ctx->capture.pending_buffers, stmpp_buffer, entry);
    stmpp_buffer_clr_pending(stmpp_buffer);

    LOGV(3, "put frame: %d fd: %d\n", stmpp_buffer->index, stmpp_buffer->fd);
    FRAME_SetID(stmpp_buffer->stmpp_frame, stmpp_buffer->index);
    FRAME_SetFD(stmpp_buffer->stmpp_frame, stmpp_buffer->fd, 0);

    if (dec->pVdecCtx->stVdecPara.bIsBufferInDecoder[stmpp_buffer->index] ==
        MPP_FALSE) {
      VDEC_ReturnOutputFrame(dec->pVdecCtx,
                             FRAME_GetBaseData(stmpp_buffer->stmpp_frame));
    }
    stmpp_buffer_clr_locked(stmpp_buffer);
  }
  pthread_mutex_unlock(&ctx->capture.queue_mutex);

  LEAVE();
}

static void stmpp_apply_info_change(struct stmpp_dec_context *dec,
                                    MppFrame *frame) {
  struct stmpp_context *ctx = dec->ctx;
  struct stmpp_video_info video_info;

  ENTER();

  memcpy((void *)&video_info, (void *)&dec->video_info, sizeof(video_info));

  video_info.width = dec->pVdecCtx->stVdecPara.nWidth;
  video_info.height = dec->pVdecCtx->stVdecPara.nHeight;
  video_info.hor_stride = dec->pVdecCtx->stVdecPara.nWidth;
  video_info.ver_stride = dec->pVdecCtx->stVdecPara.nHeight;
  video_info.size = dec->pVdecCtx->stVdecPara.nWidth *
                    dec->pVdecCtx->stVdecPara.nHeight * 3 / 2;
  video_info.valid = true;

  dec->video_info = video_info;
  dec->video_info.dirty = true;
  dec->video_info.event = dec->event_subscribed;

  //
  // Use ver_stride as new height, the visible rect would be returned
  // in g_selection.

  ctx->capture.format.width = dec->video_info.hor_stride;
  ctx->capture.format.height = dec->video_info.ver_stride;
  ctx->capture.format.num_planes = 1;
  ctx->capture.format.plane_fmt[0].bytesperline = dec->video_info.hor_stride;
  ctx->capture.format.plane_fmt[0].sizeimage = dec->video_info.size;

  LOGV(1, "frame info changed: %dx%d(%dx%d:%d)\n", dec->video_info.width,
       dec->video_info.height, dec->video_info.hor_stride,
       dec->video_info.ver_stride, dec->video_info.size);

  LEAVE();
}

static void *decoder_thread_fn(void *data) {
  struct stmpp_dec_context *dec = data;
  struct stmpp_context *ctx = dec->ctx;
  struct stmpp_buffer *stmpp_buffer;

  // MppBuffer buffer;
  int ret;
  int index;

  ENTER();

  LOGV(1, "ctx(%p): starting decoder thread\n", (void *)ctx);

  while (1) {
    pthread_mutex_lock(&dec->decoder_mutex);
    while (!dec->mpp_streaming)
      pthread_cond_wait(&dec->decoder_cond, &dec->decoder_mutex);
    //  Feed available packets and frames to mpp
    stmpp_put_packets(dec);
    stmpp_put_frames(dec);

    MppFrame *frame = FRAME_Create();
    ret = VDEC_RequestOutputFrame(dec->pVdecCtx, FRAME_GetBaseData(frame));

    pthread_mutex_unlock(&dec->decoder_mutex);

    if (ret != MPP_OK && ret != MPP_RESOLUTION_CHANGED) {
      // Handle eos frame, returning eos packet to userspace
      if (/*FRAME_GetEos(frame) || */ ret == MPP_CODER_EOS) {
        if (dec->eos_packet) {
          assert(ctx->output.streaming);

          dec->eos_packet->bytesused = 0;

          LOGV(1, "return eos packet: %d\n", dec->eos_packet->index);

          pthread_mutex_lock(&ctx->output.queue_mutex);
          TAILQ_INSERT_TAIL(&ctx->output.avail_buffers, dec->eos_packet, entry);
          stmpp_buffer_set_available(dec->eos_packet);
          pthread_mutex_unlock(&ctx->output.queue_mutex);
          dec->eos_packet = NULL;
        }
        usleep(10000);
        goto next;
      }

      if (ret == MPP_CODER_NULL_DATA) {
        VDEC_ReturnOutputFrame(dec->pVdecCtx, FRAME_GetBaseData(frame));
        goto next;
      }

      usleep(1000);
      goto next;
    }

    if (!frame) {
      usleep(1000);
      goto next;
    }

    pthread_mutex_lock(&ctx->ioctl_mutex);

    if (!dec->mpp_streaming) goto next_locked;

    //  Handle info change frame
    if (ret == MPP_RESOLUTION_CHANGED /*mpp_frame_get_info_change(frame)*/) {
      stmpp_apply_info_change(dec, frame);
      VDEC_ReturnOutputFrame(dec->pVdecCtx, FRAME_GetBaseData(frame));
      dec->pVdecCtx->stVdecPara.bIsResolutionChanged = MPP_FALSE;

      goto next_locked;
    }

    if (!ctx->capture.streaming) {
      VDEC_ReturnOutputFrame(dec->pVdecCtx, FRAME_GetBaseData(frame));
      goto next_locked;
    }

    index = FRAME_GetID(frame);
    LOGV(1, "GOOD! get a frame libv4l2, index = %d\n", index);
    stmpp_buffer = &ctx->capture.buffers[index];
    if (stmpp_buffer_pending(stmpp_buffer) ||
        stmpp_buffer_locked(stmpp_buffer)) {
      LOGE("buffer is not ready, return directly!!\n");
      VDEC_ReturnOutputFrame(dec->pVdecCtx, FRAME_GetBaseData(frame));
      goto next_locked;
    }

    stmpp_buffer->timestamp = FRAME_GetPts(frame);
    stmpp_buffer->fd = FRAME_GetFD(frame, 0);
    // LOGE("pts frame: %lld\n", stmpp_buffer->timestamp);
    FRAME_SetID(stmpp_buffer->stmpp_frame, FRAME_GetID(frame));
    // stmpp_buffer_clr_locked(stmpp_buffer);

    stmpp_buffer_set_locked(stmpp_buffer);
    /*if (mpp_frame_get_errinfo(frame) ||
        mpp_frame_get_discard(frame)) {
            LOGE("frame err or discard\n");
            stmpp_buffer->bytesused = 0;
            stmpp_buffer_set_error(stmpp_buffer);
    } else {*/
    // Size of NV12 image
    stmpp_buffer->bytesused =
        dec->video_info.hor_stride * dec->video_info.ver_stride * 3 / 2;
    //}

    LOGV(3, "return frame: %d(%" PRIu64 ")\n", index, stmpp_buffer->timestamp);

    // Report new frame to count fps
    stmpp_new_frame(ctx);
    pthread_mutex_lock(&ctx->capture.queue_mutex);

    TAILQ_INSERT_TAIL(&ctx->capture.avail_buffers, stmpp_buffer, entry);
    stmpp_buffer_set_available(stmpp_buffer);
    pthread_mutex_unlock(&ctx->capture.queue_mutex);
  next_locked:
    pthread_mutex_unlock(&ctx->ioctl_mutex);
  next:
    // Update poll event after every loop
    pthread_mutex_lock(&ctx->ioctl_mutex);
    stmpp_update_poll_event(ctx);
    pthread_mutex_unlock(&ctx->ioctl_mutex);
    FRAME_Destory(frame);
  }

  LEAVE();

  return NULL;
}

static int stmpp_dec_g_fmt(struct stmpp_dec_context *dec,
                           struct v4l2_format *f) {
  struct stmpp_context *ctx = dec->ctx;
  int ret;

  ENTER();

  LOGV(4, "f->type = %d\n", f->type);

  // The chromium expected EINVAL when resolution not available
  if (f->type == V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE && !dec->video_info.valid) {
    LOGV(1, "cannot provide resolution yet\n");
    RETURN_ERR(EINVAL, -1);
  }

  ret = stmpp_g_fmt(dec->ctx, f);

  LEAVE();
  return ret;
}

static int stmpp_dec_subscribe_event(struct stmpp_dec_context *dec,
                                     struct v4l2_event_subscription *sub,
                                     bool subscribe) {
  struct stmpp_context *ctx = dec->ctx;

  ENTER();

  if (sub->type != V4L2_EVENT_SOURCE_CHANGE) {
    LOGE("unsupported event type: %x\n", sub->type);
    RETURN_ERR(EINVAL, -1);
  }

  dec->event_subscribed = subscribe;

  LEAVE();
  return 0;
}

static int stmpp_dec_dqevent(struct stmpp_dec_context *dec,
                             struct v4l2_event *event) {
  struct stmpp_context *ctx = dec->ctx;

  ENTER();

  if (!dec->event_subscribed || !dec->video_info.event) {
    LOGV(4, "no available event\n");
    errno = ENOENT;
    return -1;
  }

  event->type = V4L2_EVENT_SOURCE_CHANGE;
  event->u.src_change.changes = V4L2_EVENT_SRC_CH_RESOLUTION;
  dec->video_info.event = false;
  LOGV(1, "dequeue resolution change event\n");

  LEAVE();
  return 0;
}

static int stmpp_dec_streamon(struct stmpp_dec_context *dec,
                              enum v4l2_buf_type *type) {
  struct stmpp_context *ctx = dec->ctx;
  struct stmpp_buf_queue *queue;

  // MppPollType poll_type;
  // uint32_t split;
  int ret;

  ENTER();

  queue = stmpp_get_queue(ctx, *type);
  if (!queue) RETURN_ERR(errno, -1);

  if (queue->streaming) goto out;
  queue->streaming = true;

  LOGV(1, "queue(%d) start streaming\n", *type);

  // Commit pending info change to mpp
  if (dec->mpp_streaming && dec->video_info.dirty &&
      *type == V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE) {
    LOGV(1, "send info change ready\n");

    dec->video_info.dirty = false;
  }

  if (dec->mpp_streaming) goto out;

  LOGV(1, "mpp start streaming\n");

  dec->pVdecCtx = VDEC_CreateChannel();
  if (!dec->pVdecCtx) {
    LOGE("Failed to create STMPP VDEC channel.\n");
    ret = ENOMEM;
    goto err;
  }

  // set para
  dec->pVdecCtx->stVdecPara.eCodingType = ctx->output.stmpp_format->type;
  dec->pVdecCtx->stVdecPara.bInputBlockModeEnable = MPP_FALSE;
  dec->pVdecCtx->stVdecPara.bOutputBlockModeEnable = MPP_FALSE;
  dec->pVdecCtx->stVdecPara.nWidth = 0;
  dec->pVdecCtx->stVdecPara.nHeight = 0;
  dec->pVdecCtx->stVdecPara.nStride = 0;
  dec->pVdecCtx->stVdecPara.nAlign = 0;
  dec->pVdecCtx->stVdecPara.eOutputPixelFormat = PIXEL_FORMAT_NV12;
  dec->pVdecCtx->stVdecPara.nScale = 1;
  dec->pVdecCtx->stVdecPara.nHorizonScaleDownRatio = 1;
  dec->pVdecCtx->stVdecPara.nVerticalScaleDownRatio = 1;
  dec->pVdecCtx->stVdecPara.nRotateDegree = 0;
  dec->pVdecCtx->stVdecPara.bThumbnailMode = 0;
  dec->pVdecCtx->stVdecPara.bIsInterlaced = MPP_FALSE;
  dec->pVdecCtx->eCodecType = CODEC_V4L2_LINLONV5V7;
  dec->pVdecCtx->stVdecPara.nInputBufferNum = 8;
  dec->pVdecCtx->stVdecPara.nOutputBufferNum = 18;
  dec->pVdecCtx->stVdecPara.eFrameBufferType =
      MPP_FRAME_BUFFERTYPE_DMABUF_INTERNAL;

  // vdec init
  ret = VDEC_Init(dec->pVdecCtx);
  if (ret != MPP_OK) {
    LOGE("Failed to initialize STMPP VDEV (ret = %d).\n", ret);
    goto err_destroy_mpp;
  }

  // Notify decoder thread to start streaming
  pthread_mutex_lock(&dec->decoder_mutex);
  dec->mpp_streaming = true;
  pthread_cond_signal(&dec->decoder_cond);
  pthread_mutex_unlock(&dec->decoder_mutex);
out:
  LEAVE();
  return 0;
err_destroy_mpp:
  VDEC_DestoryChannel(dec->pVdecCtx);
err:
  queue->streaming = false;
  RETURN_ERR(EPIPE, -1);

  return 0;
}

static int stmpp_dec_streamoff(struct stmpp_dec_context *dec,
                               enum v4l2_buf_type *type) {
  struct stmpp_context *ctx = dec->ctx;
  struct stmpp_buffer *stmpp_buffer;
  struct stmpp_buf_queue *queue;
  unsigned int i;

  ENTER();
  LOGE("xxxxxx stmpp_dec_streamoff start\n");
  queue = stmpp_get_queue(ctx, *type);
  if (!queue) RETURN_ERR(errno, -1);
  if (!queue->streaming) goto out;
  queue->streaming = false;
  LOGV(1, "queue(%d) stop streaming\n", *type);
  pthread_mutex_lock(&dec->decoder_mutex);
  // Hand over all buffers to userspace
  pthread_mutex_lock(&queue->queue_mutex);
  TAILQ_INIT(&queue->avail_buffers);
  TAILQ_INIT(&queue->pending_buffers);
  pthread_mutex_unlock(&queue->queue_mutex);
  // Update poll event after avail list changed
  stmpp_update_poll_event(ctx);

  // Reset buffer states
  for (i = 0; i < queue->num_buffers; i++) {
    stmpp_buffer = &queue->buffers[i];

    if (stmpp_buffer_error(stmpp_buffer)) stmpp_buffer_clr_error(stmpp_buffer);

    if (!stmpp_buffer_locked(stmpp_buffer)) {
      // mpp_buffer_inc_ref(stmpp_buffer->stmpp_buf);
      stmpp_buffer_set_locked(stmpp_buffer);
    }

    if (stmpp_buffer_queued(stmpp_buffer))
      stmpp_buffer_clr_queued(stmpp_buffer);

    if (stmpp_buffer_pending(stmpp_buffer))
      stmpp_buffer_clr_pending(stmpp_buffer);

    if (stmpp_buffer_available(stmpp_buffer))
      stmpp_buffer_clr_available(stmpp_buffer);
  }
  // Clear eos packet
  if (*type == V4L2_BUF_TYPE_VIDEO_OUTPUT_MPLANE) dec->eos_packet = NULL;

  // Stop mpp streaming only when all queues stopped
  if (!dec->mpp_streaming || ctx->output.streaming || ctx->capture.streaming) {
    pthread_mutex_unlock(&dec->decoder_mutex);
    goto out;
  }

  LOGV(1, "mpp stop streaming\n");
  dec->mpp_streaming = false;

  VDEC_DestoryChannel(dec->pVdecCtx);
  pthread_mutex_unlock(&dec->decoder_mutex);

out:
  LOGE("xxxxxx stmpp_dec_streamoff finish\n");
  LEAVE();
  return 0;
}

static int stmpp_dec_g_selection(struct stmpp_dec_context *dec,
                                 struct v4l2_selection *selection) {
  struct stmpp_context *ctx = dec->ctx;

  ENTER();

  if (selection->type != V4L2_BUF_TYPE_VIDEO_CAPTURE ||
      selection->target != V4L2_SEL_TGT_COMPOSE) {
    LOGE("invalid type or target\n");
    RETURN_ERR(EINVAL, -1);
  }

  // The chromium uses g_selection to get visible rect
  if (!dec->video_info.valid) {
    LOGV(1, "cannot provide visible resolution yet\n");
    RETURN_ERR(EBUSY, -1);
  }

  // Return visible rect
  selection->r.top = selection->r.left = 0;
  selection->r.width = dec->video_info.width;
  selection->r.height = dec->video_info.height;

  LOGV(1, "visible rect: %dx%d\n", selection->r.width, selection->r.height);

  LEAVE();
  return 0;
}

static int stmpp_dec_g_ctrl(struct stmpp_dec_context *dec,
                            struct v4l2_control *ctrl) {
  struct stmpp_context *ctx = dec->ctx;

  ENTER();

  if (ctrl->id != V4L2_CID_MIN_BUFFERS_FOR_CAPTURE) {
    LOGE("only support V4L2_CID_MIN_BUFFERS_FOR_CAPTURE now\n");
    RETURN_ERR(EINVAL, -1);
  }

  if (!ctx->output.stmpp_format) {
    LOGE("cannot get min buffers before s_fmt\n");
    RETURN_ERR(EBUSY, -1);
  }

  ctrl->value = dec->pVdecCtx->stVdecPara.nOutputBufferNum;
  LOGV(1, "ctrl->value  = %d\n", ctrl->value);

  // The chromium would try to require at least 5 extra buffers
  ctrl->value -= 5;

  LEAVE();
  return 0;
}

static int stmpp_dec_g_ext_ctrls(struct stmpp_dec_context *dec,
                                 struct v4l2_ext_controls *ext_ctrls) {
  struct stmpp_context *ctx = dec->ctx;
  struct v4l2_control ctrl;
  unsigned int i;

  ENTER();

  for (i = 0; i < ext_ctrls->count; i++) {
    struct v4l2_ext_control *ext_ctrl = &ext_ctrls->controls[i];

    ctrl.id = ext_ctrl->id;
    if (stmpp_dec_g_ctrl(dec, &ctrl) < 0) return -1;

    ext_ctrl->value = ctrl.value;
  }

  LEAVE();
  return 0;
}

bool stmpp_dec_has_event(void *data) {
  struct stmpp_dec_context *dec = data;

  return dec->video_info.event;
}

void *stmpp_dec_init(struct stmpp_context *ctx) {
  struct stmpp_dec_context *dec;

  ENTER();

  dec = (struct stmpp_dec_context *)calloc(1, sizeof(struct stmpp_dec_context));
  if (!dec) RETURN_ERR(ENOMEM, NULL);

  ctx->formats = stmpp_dec_fmts;
  ctx->num_formats = ARRAY_SIZE(stmpp_dec_fmts);
  dec->ctx = ctx;

  if (!ctx->max_width) ctx->max_width = 3840;

  if (!ctx->max_height) ctx->max_height = 2160;

  pthread_cond_init(&dec->decoder_cond, NULL);
  pthread_mutex_init(&dec->decoder_mutex, NULL);
  pthread_create(&dec->decoder_thread, NULL, decoder_thread_fn, dec);

  LEAVE();
  return dec;
}

void stmpp_dec_deinit(void *data) {
  struct stmpp_dec_context *dec = data;
  struct stmpp_context *ctx = dec->ctx;

  ENTER();

  if (dec->decoder_thread) {
    pthread_cancel(dec->decoder_thread);
    pthread_join(dec->decoder_thread, NULL);
  }

  if (dec->mpp_streaming) {
    // ctx->mpi->reset(ctx->mpp);
    // mpp_destroy(ctx->mpp);
  }

  free(dec);

  // LEAVE();
}

int stmpp_dec_ioctl(void *data, unsigned long cmd, void *arg) {
  struct stmpp_dec_context *dec = data;
  struct stmpp_context *ctx = dec->ctx;
  int ret;

  ENTER();

  switch (cmd) {
    /* Common ioctls */
    case VIDIOC_QUERYCAP:
      ret = stmpp_querycap(ctx, arg);
      break;
    case VIDIOC_ENUM_FMT:
      ret = stmpp_enum_fmt(ctx, arg);
      break;
    case VIDIOC_ENUM_FRAMESIZES:
      ret = stmpp_enum_framesizes(ctx, arg);
      break;
    case VIDIOC_TRY_FMT:
      ret = stmpp_try_fmt(ctx, arg);
      break;
    case VIDIOC_S_FMT:
      ret = stmpp_s_fmt(ctx, arg);
      break;
    case VIDIOC_REQBUFS:
      ret = stmpp_reqbufs(ctx, arg);
      break;
    case VIDIOC_QUERYBUF:
      ret = stmpp_querybuf(ctx, arg);
      break;
    case VIDIOC_EXPBUF:
      ret = stmpp_expbuf(ctx, arg);
      break;
    case VIDIOC_QBUF:
      ret = stmpp_qbuf(ctx, arg);

      /* Feed available packets and frames to mpp */
      if (dec->mpp_streaming) {
        stmpp_put_packets(dec);
        stmpp_put_frames(dec);
      }
      break;
    case VIDIOC_DQBUF:
      ret = stmpp_dqbuf(ctx, arg);
      break;

    /* Decoder special ioctls */
    case VIDIOC_G_FMT:
      ret = stmpp_dec_g_fmt(dec, arg);
      break;
    case VIDIOC_SUBSCRIBE_EVENT:
      ret = stmpp_dec_subscribe_event(dec, arg, true);
      break;
    case VIDIOC_UNSUBSCRIBE_EVENT:
      ret = stmpp_dec_subscribe_event(dec, arg, false);
      break;
    case VIDIOC_DQEVENT:
      ret = stmpp_dec_dqevent(dec, arg);
      break;
    case VIDIOC_STREAMON:
      ret = stmpp_dec_streamon(dec, arg);
      break;
    case VIDIOC_STREAMOFF:
      ret = stmpp_dec_streamoff(dec, arg);
      break;
    case VIDIOC_G_SELECTION:
      ret = stmpp_dec_g_selection(dec, arg);
      break;
    case VIDIOC_G_CTRL:
      ret = stmpp_dec_g_ctrl(dec, arg);
      break;
    case VIDIOC_G_EXT_CTRLS:
      ret = stmpp_dec_g_ext_ctrls(dec, arg);
      break;
    default:
      LOGE("unsupported ioctl cmd: %s(%lu)!\n", stmpp_cmd2str(cmd), cmd);
      RETURN_ERR(ENOTTY, -1);
  }

  LEAVE();
  return ret;
}
